
/*

Chroma Key (Greenscreening) Effect

The chroma key effect, also known as bluescreening or greenscreening,
sets a color or range of colors in an image to transparent (alpha = 0)
so that you can substitute a different background image.

NOTE 1: The Green Screen and Background images must be same size to avoid image clipping in output imageview.

NOTE 2: See HSV Color Range Demo to visualize calculation of HSV color ranges

Ken Shmidheiser
6 February 2020

*/

include "Tlbx CocoaUI.incl"
cocoainit

output file "Green Screen Merged Images"

include resources "greenscreen_image.jpg"
// include resources "white_screen_image.jpg"
// include resources "blue_screen_image.jpg"
include resources "background.jpg"

_window = 1

begin enum output 1
_originalImageView
_coloredDeletedView
_backgroundImageView
_mergedImageView
_originalImageLabel
_colorDeletedImageLabel
_backgroundImageLabel
_mergedImageLabel
end enum


void local fn BuildWindow
'~'1
dim as SInt32        wndStyleMask
dim as CGRect        r

wndStyleMask  = NSTitledWindowMask
wndStyleMask += NSClosableWindowMask
wndStyleMask += NSMiniaturizableWindowMask

r = fn CGRectMake( 0, 0, 1380, 360 )
cocoa window _window, @"Green Screen Merged Images", r, wndStyleMask

r = fn CGRectMake( 20, 20, 320, 300 )
imageview _originalImageView, YES,, r, NSImageScaleAxesIndependently, NSImageAlignCenter, NSImageFramePhoto, _window

r = fn CGRectMake( 360, 20, 320, 300 )
imageview _coloredDeletedView, YES,, r, NSImageScaleAxesIndependently, NSImageAlignCenter, NSImageFrameNone, _window

r = fn CGRectMake( 700, 20, 320, 300 )
imageview _backgroundImageView, YES,, r, NSImageScaleAxesIndependently, NSImageAlignCenter, NSImageFramePhoto, _window

r = fn CGRectMake( 1040, 20, 320, 300 )
imageview _mergedImageView, YES,, r, NSImageScaleAxesIndependently, NSImageAlignCenter, NSImageFramePhoto, _window

r = fn CGRectMake( 20, 330, 320, 17 )
textlabel _originalImageLabel, @"Subject Photographed Against Green Screen", r, _window
ControlSetAlignment( _originalImageLabel, NSCenterTextAlignment )

r = fn CGRectMake( 350, 330, 330, 17 )
textlabel _colorDeletedImageLabel, @"Custom ChromaKey CIFilter Sets Green Alpha to 0.0", r, _window
ControlSetAlignment( _colorDeletedImageLabel, NSCenterTextAlignment )

r = fn CGRectMake( 700, 330, 320, 17 )
textlabel _backgroundImageLabel, @"Background Image", r, _window
ControlSetAlignment( _backgroundImageLabel, NSCenterTextAlignment )

r = fn CGRectMake( 1040, 330, 320, 17 )
textlabel _mergedImageLabel, @"Filtered Image Merged with Background", r, _window
ControlSetAlignment( _mergedImageLabel, NSCenterTextAlignment )
end fn


local fn ImageRefToCIImage( image as ImageRef ) as CIImageRef
'~'1
dim as CFDataRef   dta
dim as CIImageRef  ciImage

dta = fn ImageTIFFRepresentationUsingCompression( image, NSTIFFCompressionNone, 0.0 )
ciImage = fn CIImageWithData( dta )
end fn = ciImage


local fn CIImageToImageRef( ciImage as CIImageRef ) as ImageRef
'~'1
dim as CIImageRepRef  rep
dim as CGSize         size
dim as ImageRef       image

rep = fn CIImageRepWithCIImage( ciImage )
size = fn ImageRepSize( rep )
image = fn ImageWithSize( size )
ImageAddRepresentation( image, rep )
end fn = image


// Create a custom Chroma Key CIFilter
local fn ChromaKeyFilterFromHues( minHue as CGFloat, maxHue as CGFloat ) as CIFilterRef
'~'1
dim as CIFilterRef        chromaKeyFilter
dim as CFMutableDataRef   cubeData
dim as long               x, y, z
dim as CGFloat            blue, green, red, hue, saturation, brightness
dim as ColorRef           col
dim as float              alpha, premultipliedRed, premultipliedGreen, premultipliedBlue

_size = 64
_cubeDataSize = _size * _size * _size * 4

chromaKeyFilter = fn CIFilterWithName( @"CIColorCube" )

cubeData = fn DataWithCapacity( _cubeDataSize * sizeof(float) )

for z = 0 to _size -1
blue = ((double)z)/(_size-1)

for y = 0 to _size -1
green = ((double)y)/(_size-1)

for x = 0 to _size -1
red = ((double)x)/(_size-1)

col = fn ColorWithRGB( red, green, blue, 1.0 )

/*
The next three lines convert RGB to HSV (hue-saturation-value)
or HSB (hue-saturation-brightness in CocoaUI)
for calculation of cubeData values
*/
ColorGetHSB( col, @hue, @saturation, @brightness, NULL )
alpha = ( hue >= minHue && hue <= maxHue )
if alpha == 0 then alpha = 1

premultipliedRed   = red   * alpha
premultipliedGreen = green * alpha
premultipliedBlue  = blue  * alpha

DataAppendBytes( cubeData, @premultipliedRed,   sizeof(float) )
DataAppendBytes( cubeData, @premultipliedGreen, sizeof(float) )
DataAppendBytes( cubeData, @premultipliedBlue,  sizeof(float) )
DataAppendBytes( cubeData, @alpha,              sizeof(float) )
next
next
next
ObjectSetValueForkey( chromaKeyFilter, @(_size), @"inputCubeDimension" )
ObjectSetValueForkey( chromaKeyFilter, cubeData, @"inputCubeData" )
end fn = chromaKeyFilter


local fn RemoveColorFromSource( foregroundCIImage as CIImageRef, minHue as CGFloat, maxHue as CGFloat ) as CIImageRef
'~'1
dim as CIFilterRef chromaKeyFilter
dim as CIImageRef  sourceCIImageWithoutBackground

// Apply custom ChromaKey CIFilter
chromaKeyFilter = fn ChromaKeyFilterFromHues( minHue, maxHue )
ObjectSetValueForkey( chromaKeyFilter, foregroundCIImage, @"inputImage" )
sourceCIImageWithoutBackground = fn CIFilterOutputImage( chromaKeyFilter )
end fn = sourceCIImageWithoutBackground


local fn MergeImages( sourceCIImageWithoutBackground as CIImageRef, backgroundCIImage as CIImageRef ) as CIImageRef
'~'1
dim as CIImageRef  compositedCIImage

compositedCIImage = fn CIImageByCompositingOverImage( sourceCIImageWithoutBackground, backgroundCIImage )
end fn = compositedCIImage


void local fn BuildDemoViews
'~'1
dim as ImageRef   foregroundImage, colorDeletedImage, backgroundImage, mergedImage
dim as CIImageRef ciForeground, ciBackground, ciFixedForeground, ciMergedImage

foregroundImage = fn ImageNamed( @"greenscreen_image.jpg" )
backgroundImage = fn ImageNamed( @"background.jpg" )

ImageViewSetImage( _originalImageView,   foregroundImage )
ImageViewSetImage( _backgroundImageView, backgroundImage )

ciForeground = fn ImageRefToCIImage( foregroundImage )
ciBackground = fn ImageRefToCIImage( backgroundImage )

/*
0.30, 0.40 HSV angles filter out green screen
0.00, 0.00 fllters out white screen
0.51, 0.68 filters out blue screen

Green screens are preferred since they don't affect human skin tones.

For formula to determine other color-filter values see:
https://developer.apple.com/documentation/coreimage/applying_a_chroma_key_effect?language=objc
*/
ciFixedForeground = fn RemoveColorFromSource( ciForeground, 0.30, 0.40 )
colorDeletedImage = fn CIImageToImageRef( ciFixedForeground )
ImageViewSetImage( _coloredDeletedView, colorDeletedImage )

ciMergedImage = fn MergeImages( ciFixedForeground, ciBackground )

mergedImage = fn CIImageToImageRef( ciMergedImage )
ImageViewSetImage( _mergedImageView, mergedImage )
end fn

fn BuildWindow
fn BuildDemoViews

HandleEvents

