
/*

HSV Color Range Demo

The Core Imqge CIColorCube filter -- which is key to creating a chromakey green screen application
requires HSV (Hue, Saturation, Value) colors rather than ColorRefs. The HSV model represents color in three dimensions.
In these models, colors of each hue are arranged in a radial slice, around a central axis of neutral colors
which range from black at the bottom to white at the top. The model is shaped like an inverted cone with the
flat top looking like a standard, fully-saturated RGB color wheel with a white dot in
the center where the colors converge, tapering down to the black at the iverted cone's point.

In FB's CocoaUI, the Color header contains funtions to convert ColorRefs to and from HSB (Hue, Saturating, Brightness)
colors which can be used to represent HSV colors in two diemnsions. 

A green screen chromakey is used to eliminate green backgrounds behind an object. It does this by reducing the green
color's alpha value to 0 giving the subject a transparent background. The subject can then be overlaid onto another
background. An example is the typical television weather report where the weather forecaster stands before a green
screen, the green background is eliminated and his/her image is superimposed on a weather map. A problem is that the
filter will make any shade of green in the image transparent, so if the forecaster is wearing a green shirt or blouse, it
will become tranparent along with the background.

In programming a green screen chromakey demo in FB, I created this utility to allow me to visulaize how HSV hue angle
and destination angles work. The demo opens with values set to eliminate standard green screen hues (minimum hue angle
value of 0.30 and maximum hue angle value of 0.40.

Note that backgrounds of other colors can also be removed by changing the HSV values. However, green is the standard
because it does not intefere with human skin tones that contain red.

This demo does contain a bug that allows it to display converted minimum and maximum hue values outside the 0 to 1
range. But it does offer a rough visual approximation of what hues of color will be removed from an image.

Ken Shmidheiser
5 February 2020

*/

include "Tlbx CocoaUI.incl"
cocoainit

output file "HSV Color Range"

include resources "color_wheel.jpg"

_window = 1

begin enum output 1
_originalImageView
_colorWheelImageView
_centerHueAngleSlider
_centerHueValueField
_colorAngleLabel
_degreesSeparationSlider
_degreesSeparationValue
_degresSeparationLabel
_angleValues
end enum


void local fn BuildWindow
'~'1
dim as SInt32        wndStyleMask
dim as CGRect        r

wndStyleMask  = NSTitledWindowMask
wndStyleMask += NSClosableWindowMask
wndStyleMask += NSMiniaturizableWindowMask

r = fn CGRectMake( 0, 0, 1075, 430 )
cocoa window _window, @"HSV Color Range", r, wndStyleMask

r = fn CGRectMake( 10, 10, 400, 400 )
imageview _originalImageView, YES, @"color_wheel.jpg", r, NSImageScaleAxesIndependently, NSImageAlignCenter, NSImageFramePhoto, _window

r = fn CGRectMake( 420, 10, 400, 400 )
imageview _colorWheelImageView, YES,, r, NSImageScaleAxesIndependently, NSImageAlignCenter, NSImageFramePhoto, _window

r = fn CGRectMake( 858, 30, 20, 350 )
slider _centerHueAngleSlider, YES, 126.51, r, 0.00, 360.00, NO, _window
ControlSetContinuous( _centerHueAngleSlider, YES )

r = fn CGRectMake( 838, 10, 50, 16 )
textlabel _centerHueValueField, @"126.51", r, _window
ControlSetAlignment( _centerHueValueField, NSCenterTextAlignment )

r = fn CGRectMake( 836, 382, 60, 35 )
textlabel _colorAngleLabel, @"Color Angleº", r, _window
ControlSetAlignment( _colorAngleLabel, NSCenterTextAlignment )

r = fn CGRectMake( 908, 30, 20, 350 )
slider _degreesSeparationSlider, YES, 37.096, r, 0.00, 120.00, NO, _window
ControlSetContinuous( _degreesSeparationSlider, YES )

r = fn CGRectMake( 890, 10, 50, 16 )
textlabel _degreesSeparationValue, @"37.069", r, _window
ControlSetAlignment( _degreesSeparationValue, NSCenterTextAlignment )

r = fn CGRectMake( 885, 382, 70, 35 )
textlabel _degresSeparationLabel, @"Degrees Range", r, _window
ControlSetAlignment( _degresSeparationLabel, NSCenterTextAlignment )

r = fn CGRectMake( 965, 335, 90, 80 )
textfield _angleValues, YES, , r, _window
TextFieldSetBezelStyle( _angleValues, NSTextFieldSquareBezel )
end fn


local fn ImageRefToCIImage( image as ImageRef ) as CIImageRef
'~'1
dim as CFDataRef   dta
dim as CIImageRef  ciImage

dta = fn ImageTIFFRepresentationUsingCompression( image, NSTIFFCompressionNone, 0.0 )
ciImage = fn CIImageWithData( dta )
end fn = ciImage


local fn CIImageToImageRef( ciImage as CIImageRef ) as ImageRef
'~'1
dim as CIImageRepRef  rep
dim as CGSize         size
dim as ImageRef       image

rep = fn CIImageRepWithCIImage( ciImage )
size = fn ImageRepSize( rep )
image = fn ImageWithSize( size )
ImageAddRepresentation( image, rep )
end fn = image


// Create a custom ChromaKey CIFilter
local fn ChromaKeyFilterFromHues( centerHueAngle as double, degreeRange as double ) as CIFilterRef
'~'1
dim as CIFilterRef        chromaKeyFilter
dim as CFMutableDataRef   cubeData
dim as long               x, y, z
dim as CGFloat            minHueAngle, maxHueAngle
dim as CGFloat            blue, green, red, hue, saturation, brightness
dim as ColorRef           col
dim as float              alpha, premultipliedRed, premultipliedGreen, premultipliedBlue
dim as CFStringRef        hueAnglesStr

minHueAngle = (centerHueAngle - degreeRange/2.0) / 360 // 60 degree range = +30 -30
maxHueAngle = (centerHueAngle + degreeRange/2.0) / 360

hueAnglesStr =  fn StringWithFormat( @"Min. Hue Angle: %0.2f\nMax. Hue Angle: %0.2f", minHueAngle, maxHueAngle )
ControlSetStringValue( _angleValues, hueAnglesStr )

_size = 64
_cubeDataSize = _size * _size * _size * 4

chromaKeyFilter = fn CIFilterWithName( @"CIColorCube" )

cubeData = fn DataWithCapacity( _cubeDataSize * sizeof(float) )

for z = 0 to _size -1
blue = ((double)z)/(_size-1)

for y = 0 to _size -1
green = ((double)y)/(_size-1)

for x = 0 to _size -1
red = ((double)x)/(_size-1)

col = fn ColorWithRGB( red, green, blue, 1.0 )

// Next three lines convert RGB to HSV (or HSB — hue-saturation-brightness in CocoaUI) for calculation of cubeData values
ColorGetHSB( col, @hue, @saturation, @brightness, NULL )
alpha = ( hue >= minHueAngle && hue <= maxHueAngle )
if alpha == 0 then alpha = 1

premultipliedRed   = red   * alpha
premultipliedGreen = green * alpha
premultipliedBlue  = blue  * alpha

DataAppendBytes( cubeData, @premultipliedRed,   sizeof(float) )
DataAppendBytes( cubeData, @premultipliedGreen, sizeof(float) )
DataAppendBytes( cubeData, @premultipliedBlue,  sizeof(float) )
DataAppendBytes( cubeData, @alpha,              sizeof(float) )
next
next
next
ObjectSetValueForkey( chromaKeyFilter, @(_size), @"inputCubeDimension" )
ObjectSetValueForkey( chromaKeyFilter, cubeData, @"inputCubeData" )
end fn = chromaKeyFilter


local fn RemoveColorFromSource( foregroundCIImage as CIImageRef, centerHueAngle as double, degreeRange as double ) as CIImageRef
'~'1
dim as CIFilterRef chromaKeyFilter
dim as CIImageRef  sourceCIImageWithoutBackground

// Apply custom Chroma Key CIFilter
chromaKeyFilter = fn ChromaKeyFilterFromHues( centerHueAngle, degreeRange )
ObjectSetValueForkey( chromaKeyFilter, foregroundCIImage, @"inputImage" )
sourceCIImageWithoutBackground = fn CIFilterOutputImage( chromaKeyFilter )
end fn = sourceCIImageWithoutBackground


void local fn BuildColorDeleteImage
'~'1
dim as ImageRef    foregroundImage, colorDeletedImage
dim as float       centerHue, degreeRange
dim as CIImageRef  ciForeground, ciFixedForeground


foregroundImage = fn ImageNamed( @"color_wheel.jpg" )
centerHue =  fn ControlDoubleValue( _centerHueAngleSlider )
degreeRange = fn ControlDoubleValue( _degreesSeparationSlider )

ImageViewSetImage( _originalImageView, foregroundImage )
ciForeground = fn ImageRefToCIImage( foregroundImage )

/*
0.30, 0.40 filters out green screen
0.00, 0.00 fllters out white screen
0.51, 0.68 filters out blue screen

Green screens are preferred since they don't affect human skin tones.

For formula to determine other color-filter values see:
https://developer.apple.com/documentation/coreimage/applying_a_chroma_key_effect?language=objc
*/
ciFixedForeground = fn RemoveColorFromSource( ciForeground, centerHue, degreeRange )
colorDeletedImage = fn CIImageToImageRef( ciFixedForeground )
ImageViewSetImage( _colorWheelImageView, colorDeletedImage )
end fn


void local fn DoDialog( ev as long, tag as long, wnd as long, obj as CFTypeRef )
'~'1
dim as double centerAngle, degreeRange

select (ev)
case _btnClick
select (tag )
case _centerHueAngleSlider
centerAngle = fn ControlDoubleValue( tag )
ControlSetDoubleValue( _centerHueValueField, centerAngle )
fn BuildColorDeleteImage
case _degreesSeparationSlider
degreeRange = fn ControlDoubleValue( tag )
ControlSetDoubleValue( _degreesSeparationValue, degreeRange )
fn BuildColorDeleteImage
end select

case _windowWillClose : end
end select
end fn

on dialog fn DoDialog

fn BuildWindow
fn BuildColorDeleteImage

HandleEvents

