/*
     Scenario: My shareware saves, among other data, reminders established by the user. The user is then visually and audibly reminded
               ( based on dates/times they choose ) if they are at their computer. Some users travel with a 2nd computer ( or without
               a laptop ) and don't have access to their desktop computer where my shareware would remind them. One solution is to have
               the shareware app send an email to the user. The user then reads the reminder via the email from their cell phone, laptop etc ( maybe via webmail )

  Description: Sending an email from your own code without any user interaction should be simple but there are several stumbling blocks including ISPs that are 
               protecting everyone from mail spammers. One solution is to use the user's own mail app since it already has access to the SMTP server. 

        To Do: (1) This is hard-wired to Apple's Mail but it launches the default mail program. Code to handle any mail app would be more flexible.
               (2) Script would be better as an external file run by osascript. This would allow separate maintenance of the script without changing program code.
                   It would also be possible to use CFStrings to allow scripts larger than 255 characters
               (3) In a real application the email data ( subject, To etc. ) would be collected from user as CFStrings and passed directly to SendMailCF. This
                   avoids the size limitations of pascal strings. 
         
  Bugs/Limits: (1) Scripts larger than 255 characters won't work - see comments in To Do.
               (2) Email data limited to 255 characters - see comments in To Do
               (3) Pressing "send" can fail if the Mail app is not already open when the code executes. If Mail has to be launched, 
                   windows open when Mail last quit are re-opened, and this interferes with the new window created by this code.

        
                
Brian  - original demo                                 {20081209}
Bernie - cleanup, packaging and escaping url strings   {20081210}
Brian  - check for assistive devices enabled           {20081210}

Brian/Bernie  - uses Applescript from bundle           {20100622}
     • Addresses some of the 'To Do" and Bug/Limits as noted above from original LaunchMailPressSend
     • Instead of using FB's Applescript support, the code creates a unix pipe
       and opens an apple script file stored in the app bundle ( Contents/Resources )
     • The unix pipe method does not have the pascal string limitations of  FB's Open "Unix"
*/

include "Tlbx LSOpen.incl"
include "Tlbx Dialogs.incl"
include resources "myMailAscrpt.scpt"
_myReplyBufferSize   = 4096

//HIServices/AXUIElement.h
toolbox fn AXAPIEnabled() = Boolean
toolbox fn fread( pointer data, long size, long count, pointer stream ) = unsigned long

begin enum
_kCFURLCreationErr = -32022
_kCFStringCreateWithFormatErr = -32023
_kCFURLCreateStringByAddingPercentEscapesErr = -32024
end enum


// for the parms passed in to ShowAlert: passing in 0 ( zero ) for parms not used effective removes extra buttons
// Do NOT pass @"" for unused parms, pass 0
local mode
local fn ShowAlert( alertType as AlertType, message as CFStringRef, explanation as CFStringRef, defaultBtnText as CFStringRef, cancelBtnText as CFStringRef, otherBtnText as CFStringRef )
'~'1
dim as AlertStdCFStringAlertParamRec      paramRec
dim as DialogRef                          dialogRef
dim as DialogItemIndex                    itemHit

fn GetStandardAlertDefaultParams( paramRec, _kStdCFStringAlertVersionOne )
paramRec.defaultText   = defaultBtnText
paramRec.defaultButton = _kAlertStdAlertOKButton
paramRec.cancelText    = cancelBtnText
paramRec.cancelButton  = _kAlertStdAlertCancelButton
paramRec.otherText     = otherBtnText
fn CreateStandardAlert( alertType, message, explanation, @paramRec, @dialogRef )
fn RunStandardAlert( dialogRef, 0, @itemHit )
end fn = itemHit

local fn PipeOpen( cmdString as CFStringRef ) as pointer
'~'1
dim as pointer        buffer, unixPipe : unixPipe = 0
dim as CFIndex        length
dim type.2

length = fn CFStringGetMaximumSizeForEncoding( fn CFStringGetLength( cmdString ), _kCFStringEncodingUTF8 ) + 1
buffer = fn malloc( length )

type.0`` = _"r" : type.1`` = 0 // null-terminated C string

long if ( buffer ) 
long if ( fn CFStringGetCString( cmdString, #buffer, length, _kCFStringEncodingUTF8 ) )
unixPipe = fn popen( #buffer, type )
end if
end if

free( buffer )

end fn = unixPipe

local fn CreateCFStringResult( unixPipe as pointer ) as CFStringRef
'~'1
dim as pointer       p
dim as long          nChars
dim as CFStringRef   string

p = fn malloc( _myReplyBufferSize )

nChars = fn fread( #p, 1, _myReplyBufferSize, unixPipe )
fn pclose( unixPipe ) 

string  = fn CFStringCreateWithBytes( _kCFAllocatorDefault, #p, nChars, _kCFStringEncodingUTF8, _false )
free( p )
end fn = string


local fn IsAppRunning( appName as CFStringRef )
'~'1
dim as CFStringRef          resultStr
dim as pointer              unixPipe
dim as CFRange              range : range.length = 0
dim as CFOptionFlags        options : options = _kCFCompareCaseInsensitive

unixPipe = fn PipeOpen( @"ps ax -o command -c" )
long if ( unixPipe )
resultStr = fn CreateCFStringResult( #unixPipe )
range = fn CFStringFind( resultStr, appName, options )
CFRelease( resultStr )
end if
end fn = ( range.length > 0 )


local fn CreatePathToFile( fileName as CFStringRef ) as CFStringRef
'~'1
dim as CFStringRef          path
dim as CFURLRef             url

url    = fn CFBundleCopyResourceURL( fn CFBundleGetMainBundle(), fileName, 0, 0 )
path   = fn CFURLCopyFileSystemPath ( url, _kCFURLPOSIXPathStyle )
CFRelease( url ) 

end fn = path


local fn BackSlashEscapePath( mutStr as CFMutableStringRef )
'~'1
// In the replace string, two "\\" backslashes are required in order to insert one backslash
// F/R must be done for backslash first. Otherwise, work by other F/R would be destroyed

fn CFStringFindAndReplace( mutStr, @"\\",@"\\\\" , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @" ", @"\\ "  , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @"""", @"\\""", fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @"'", @"\\'"  , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @"?", @"\\?"  , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @"(", @"\\("  , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @")", @"\\)"  , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @"&", @"\\&"  , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @";", @"\\;"  , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @"|", @"\\|"  , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @"$", @"\\$"  , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @">", @"\\>"  , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
fn CFStringFindAndReplace( mutStr, @"<", @"\\<"  , fn CFRangeMake( 0, fn CFStringGetLength( mutStr ) ), 0 )
end fn



local fn SendMail( mailto as CFStringRef, cc as CFStringRef, subject as CFStringRef, message as CFStringRef )
'~'1
dim as CFURLRef             url
dim as CFStringRef                        tempString, string
dim as OSStatus                           err : err = _noErr
dim as CFStringRef                        path
dim as CFMutableStringRef                 mutStr
dim as pointer                            unixPipe

// construct url string
tempString = fn CFStringCreateWithFormat( _kCFAllocatorDefault, 0, fn CFSTR( "mailTo:%@?cc=%@&subject=%@&body=%@" ), mailto, cc, subject, message )
if ( tempString == 0 ) then err = _kCFStringCreateWithFormatErr : exit fn

// fix special characters
string = fn CFURLCreateStringByAddingPercentEscapes( _kCFAllocatorDefault, tempString, 0, 0, _kCFStringEncodingUTF8 )
CFRelease( tempString )
if ( string == 0 ) then err = _kCFURLCreateStringByAddingPercentEscapesErr : exit fn

// launch the url
url = fn CFURLCreateWithString( _kCFAllocatorDefault, string, 0 )
CFRelease( string )
if ( url == 0 ) then err = _kCFURLCreationErr : exit fn
fn LSOpenCFURLRef( url, NULL )
CFRelease( url )

long if ( fn IsAppRunning( @"Mail" ) )
path      = fn CreatePathToFile( @"myMailAscrpt.scpt" )
mutStr    = fn CFStringCreateMutableCopy( _kCFAllocatorDefault, 0, path )
CFRelease( path )
fn BackSlashEscapePath( mutStr )
string    = fn CFStringCreateWithFormat( _kCFAllocatorDefault, NULL, @"%@ %@", @"osascript", mutStr )

CFRelease( mutStr )

unixPipe  = fn PipeOpen( string )
// fn GetCFStringResult could optionally be called to retrieve result of unix call.
// Instead here we only do the fn plcose.
fn pclose( unixPipe )
CFRelease( string )
end if 
end fn = err


// Make sure the system preference Access for Assistive Devices is turned ON - required for applescript to drive GUI.
// Return value from fn ShowAlert ignored intentially for this usage.
long if ( fn AXAPIEnabled() )
//  params: to, cc, subject, message
fn SendMail( @"theUsersMail@domain.com", @"noOne@nowhere.org", @"Test Subject", @"Hello everyone" )
xelse
fn ShowAlert( _kAlertNoteAlert, @"Access for Assistive Devices must be enabled", @"Refer to Universal Access in System Preferences", @"OK", 0, 0 )
end if


