'---------------------
/*
Clickable URL, using the tracking method of Apple's QA1380.
Robert P.   April 2007

Installing a handler for the appropriate CarbonEvents has 
advantages over old programming techniques for clickable URLs
- a click on the link is tracked, and can therefore be 
  cancelled by moving the mouse away before releasing the button.
- any number of clickable URLs can be set up without change to 
  the rest of the program. A call to MakeClickableURL and you're done.
- nothing needs disposing when the control or window is closed.

This version
- allows link text to be different from the URL.
- displays the link in red while it is tracked.
- shows how to specify the font/size of the link.
- changes cursor when over link (in 10.4 or later).
- works for compositing and non-compositing windows.

- is translatable by FBtoC 1.0a89  June 2007
*/


include "Tlbx HIView.incl"
include "Tlbx CoreGraphics.incl"
include "Tlbx ControlDefinitions.incl"
end globals

_kMyURLTextTag = _"URLT"

// we get here when any link is clicked
local fn DoURLClick( c as ControlRef )
'~'1
dim as Str255  url

// retrieve the url string from the control
fn GetControlProperty( c, 0, _kMyURLTextTag, 256, #0, @url )
// The 'open' command is a handy way to open URLs as well as files
open "unix", 222, "open " + url
close 222
end fn


local fn ClickableURLHandler( nextHandler as EventHandlerCallRef, theEvent as EventRef, userData as ptr )
'~'1
dim as OSStatus             result
dim as ControlRef           c
dim as HIRect               bounds
dim as HIPoint              pt
dim as ControlPartCode    @ part
dim as ControlFontStyleRec  cfs

result = _eventNotHandledErr
c = userData
select fn GetEventKind( theEvent )

case _kEventControlHitTest
// static text control requires this handler in order to track the mouse
fn HIViewGetBounds( c, @bounds )
fn GetEventParameter( theEvent, _kEventParamMouseLocation, _typeHIPoint, #0, sizeof( pt ), #0, @pt )
part = _kControlNoPart // mouse not over control

// if mouse over the control, send kControlButtonPart
if ( fn CGRectContainsPoint( bounds, pt ) ) then part = _kControlButtonPart

fn SetEventParameter( theEvent, _kEventParamControlPart, _typeControlPartCode, sizeof( part ), @part )
result = _noErr // we handled the event

case _kEventControlHiliteChanged
// needed for colour change on click, if control is in compositing window
fn HIViewSetNeedsDisplay( c, _true )

case _kEventControlDraw
// Draw the link text either visibly-clickable (blue) or pressed (red).
// retrieve the existing ControlFontStyleRec
fn GetControlData( c, _kControlEditTextPart, _kControlStaticTextStyleTag, sizeof( cfs ), @cfs, #0 )
// set the foreColor field
long if ( fn GetControlHilite( c ) == _kControlButtonPart )
cfs.foreColor.red = 0xffff // pressed link red
cfs.foreColor.blue = 0
xelse
cfs.foreColor.red = 0
cfs.foreColor.blue = 0xeeee // special "url blue"
end if
cfs.foreColor.green = 0
fn SetControlFontStyle( c, cfs )
// let the default handler (i.e. The System) draw the text
result = fn CallNextEventHandler( nextHandler, theEvent )

case _kEventControlTrackingAreaEntered
// cursor change
if ( fn IsControlActive( c ) ) then fn SetThemeCursor( _kThemePointingHandCursor )

case _kEventControlTrackingAreaExited
// cursor change
if ( fn IsControlActive( c ) ) then fn SetThemeCursor( _kThemeArrowCursor )

case _kEventControlHit
// act on a completed click-and-release, like a button click
fn DoURLClick( c )
result = _noErr // we handled the event

end select
end fn = result


local fn InstallClickableURLHandler( c as ControlRef )
'~'1
dim as EventTypeSpec  myEvents(5)
dim as long           nEvents

// The events we intercept are all for the control
// itself (not its window or other enclosing HIViews).
nEvents = 4
myEvents.eventClass(0) = _kEventClassControl
myEvents.eventKind(0) = _kEventControlHitTest
myEvents.eventClass(1) = _kEventClassControl
myEvents.eventKind(1) = _kEventControlHit
myEvents.eventClass(2) = _kEventClassControl
myEvents.eventKind(2) = _kEventControlDraw
myEvents.eventClass(3) = _kEventClassControl
myEvents.eventKind(3) = _kEventControlHiliteChanged

/*
Prepare events and TrackingArea for cursor change.
This refinement requires 10.4 or later.
On older systems the cursor simply won't change. 
*/
long if ( system( _sysVers ) >= 1040 )
nEvents = 6
myEvents.eventClass(4) = _kEventClassControl
myEvents.eventKind(4) = _kEventControlTrackingAreaEntered
myEvents.eventClass(5) = _kEventClassControl
myEvents.eventKind(5) = _kEventControlTrackingAreaExited
fn HIViewNewTrackingArea( c, 0, 0, #0 )
end if

// install the handler with the array of events, passing the ControlRef c as userData
fn InstallEventHandler( fn GetControlEventTarget( c ), @fn ClickableURLHandler, nEvents, @myEvents(0), #c, #0 )
end fn


local fn MakeClickableURL( btnNum as long, linkTxt as Str255, url as Str255, r as ^Rect )
'~'1
dim as Rect                rCopy
dim as ControlFontStyleRec cfs
dim as ControlRef          c

rCopy = r
// make the text control
appearance button btnNum,,,,,, @rCopy, _kControlStaticTextProc

cfs.flags = _kControlUseFaceMask + _kControlUseForeColorMask
cfs.style = _ulineBit% // underlined
// uncomment to specify font & size (Courier 12 as a demo)
/*
cfs.flags = cfs.flags || _kControlUseFontMask || _kControlUseSizeMask
cfs.size = 12
cfs.font = _courier
*/
c = button&( btnNum )
fn SetControlFontStyle( c, cfs )
fn SetButtonTextString( btnNum, linkTxt )
fn SetControlProperty( c, 0, _kMyURLTextTag, 256, @url )
// install Carbon event handler on the control
fn InstallClickableURLHandler( c )
end fn


local fn SetUpMyURLs
'~'1
dim as Str255 url
dim as Rect   r

SetRect( r, 20, 15, 380, 35 )
url = "http://developer.apple.com/qa/qa2004/qa1380.html"
fn MakeClickableURL( 1, url, url, r ) // link text and URL are the same

SetRect( r, 20, 55, 130, 75 )
fn MakeClickableURL( 2, "reply to author", "mailto:nobody@nowhere.moc", r )

SetRect( r, 20, 95, 110, 115 )
fn MakeClickableURL( 3, "ftp to Apple", "ftp://ftp.apple.com", r )
end fn


appearance window 2, @"Clickable URLs (compositing window)",, _kDocumentWindowClass, _kWindowStandardHandlerAttribute + _kWindowCompositingAttribute
fn SetUpMyURLs

do
HandleEvents
until ( gFBQuit )
'---------------------

