
int static inline FileOf( int sq ) { return sq & 7; }
int static inline RankOf( int sq ) { return sq >> 3; }


/*
 Time-critical BitBoard utilities
  
 KountBits() †
 GetLowestBit() †
 GetHighestBit() †
 SetBit()
 ClearBit()
 ClearLowestBit()
 
 All are executed so frequently that overhead of a function call is noticeable. A simple and effective 
 optimization is to make them static inline functions, the modern replacement for function macros.
 
 †  No efficient high-level language implementation is possible for these functions, which
 therefore resort to assembler or, less happily, compiler builtins.
 */


#if TARGET_RT_LITTLE_ENDIAN && __LP64__
// Intel 64-bit asm


#ifdef HAS_POPCNT
// Core i5 and i7 processors have popcnt instruction

int static inline KountBits( BitBoard b )
{
	BitBoard  result, tmp;
	
	asm( "   popcnt  %1, %0  \n\t"
		 :  "=&r" (result), "=&r" (tmp)
		 :  "1" (b)
		 :  "cc" );
	return result;
}

#else

int static inline KountBits( BitBoard b )
{
	BitBoard  result, tmp1, tmp2;
	
	asm( "    xorq    %0, %0    \n\t"
		 "    testq   %1, %1    \n\t"
		 "    jz      2f        \n\t"
		 "1:  leaq    -1(%1),%2 \n\t"
		 "    incq    %0        \n\t"
		 "    andq    %2, %1    \n\t"
		 "    jnz     1b        \n\t"
		 "2:                    \n\t"
		 :   "=&r" (result), "=&r" (tmp1), "=&r" (tmp2)
		 :   "1" (b)
		 :   "cc" );
	return (int)result;
}

#endif /* HAS_POPCNT */


int static inline GetLowestBit( BitBoard b ) // LS1B
{
	BitBoard  result, tmp;
	
	asm( " bsfq  %1, %0  \n\t"
		 : "=&r" (result), "=&r" (tmp)
		 : "1" (b)
		 : "cc" );
	return (int)result;
}


int static inline GetHighestBit( BitBoard b ) // MS1B
{
	BitBoard  result, tmp;
	
	asm( " bsrq  %1, %0  \n\t"
		 : "=&r" (result), "=&r" (tmp)
		 : "1" (b)
		 : "cc" );
	return (int)result;
}

#else

static inline int KountBits( BitBoard b ) { return __builtin_popcountll( b ); } 

static inline int GetLowestBit( BitBoard b ) { return __builtin_ctzll( b ); } 

static inline int GetHighestBit( BitBoard b ) { return 63 - __builtin_clzll( b ); } 

#endif  /* TARGET_RT_LITTLE_ENDIAN && __LP64__ */




static inline void SetBit( BitBoard * b, int bitNum )
{
	*b |= (1ULL << bitNum);
} 


static inline void ClearBit( BitBoard * b, int bitNum )
{
	*b &= ~(1ULL << bitNum);
} 

// specialised faster version of ClearBit()
static inline void ClearLowestBit( BitBoard * b )
{
	*b &= *b - 1ULL;
} 

